# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from ._pattern import (Pattern, FileFromPattern, XMLFromPattern,
    XMLFromMarkupPattern, error,
    ArbitraryExtensionFileFromPattern, URLFromPattern)


URLFromPattern
ArbitraryExtensionFileFromPattern
FileFromPattern
Pattern
XMLFromMarkupPattern
XMLFromPattern
error
